//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsResSurfRevLib.h"

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions "setupShaders", "init", "initVSL", "changeSize" and "renderScene".
// 2) Follow the comments in shaders to implement the shadow mapping.
// 3) Move the light position further away by modifying "lightDist" constant, notice the loss of quality.
// 4) Increase the shadowMapSize, notice the increase in quality.
// 5) Make the object rotate around x-axis, notice the self-shadowing.
// 6) Make the object rotate around xyz-axis, enjoy the show ;-)
// 7) Change glPolygonOffset to 0.f and 100.f.
// 8) Use ortographic projection matrix - we are using directional light.
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader, shaderShadow;
VSResSurfRevLib torusCaster, sphereReceiver;

// Double buffered position texture
GLuint shadowMap;

// Double buffered texture framebuffer
GLuint shadowFramebuffer;

// Matrix for transformation from world space to shadow space
float shadowMatrix[16];

// Matrix for shadow projection
float shadowProjectionMatrix[16];

// Matrix for camera projection
float cameraProjectionMatrix[16];

// Matrix for remapping [-1,1] clip coordinates into [0,1] texture coordinates
float remapMatrix[] = {0.5f, 0.0f, 0.0f, 0.0f, 0.0f, 0.5f, 0.0f, 0.0f, 0.0f, 0.0f, 0.5f, 0.0f, 0.5f, 0.5f, 0.5f, 1.0f};

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.0f;

// Geometry offset
float offsetX = 0.f, offsetY = 0.f;

// Time counter
float time = 0.f;

// Size of the shadow texture - higher means better quality
const GLuint shadowMapSize = 1024;
const float lightDist = 3.f;

// Light direction
float lightDir[4] = {1.0f, 1.0f, 1.0f, 0.0f};


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Compute the base of the shadow matrix as projection multiplied with bias
	vsml->loadIdentity(VSMathLib::PROJECTION);
#if 1
	ratio = (1.0f * shadowMapSize) / shadowMapSize; // 1.f
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
#else
	vsml->ortho(-2.f, 2.f, -2.f, 2.f, 0.1f, 10000.0f);
#endif
	memcpy(shadowProjectionMatrix, vsml->get(VSMathLib::PROJECTION), sizeof(shadowProjectionMatrix)); // Save to global variable

	// Prepare the shadowMatrix as the remapping matrix multiplied by the projection matrix
	vsml->loadIdentity(VSMathLib::AUX0);
	vsml->multMatrix(VSMathLib::AUX0, remapMatrix);
	vsml->multMatrix(VSMathLib::AUX0, vsml->get(VSMathLib::PROJECTION));
	memcpy(shadowMatrix, vsml->get(VSMathLib::AUX0), sizeof(shadowMatrix));  // Save to global variable

	// ------------------------------------------------------------

	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
	memcpy(cameraProjectionMatrix, vsml->get(VSMathLib::PROJECTION), sizeof(cameraProjectionMatrix)); // Save to global variable
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void)
{
	// No texture
	//glBindTexture(GL_TEXTURE_2D, 0);

	// Save current viewport
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	// Bind to the writing buffer
	glBindFramebuffer(GL_FRAMEBUFFER, shadowFramebuffer);

	// Set the viewport to the texture size
	glViewport(0, 0, shadowMapSize, shadowMapSize);

	// Disable writing of color information
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera from the light - directional light
	vsml->lookAt(lightDir[0]*lightDist, lightDir[1]*lightDist, lightDir[2]*lightDist, 0,0,0, 0,1,0);

	vsml->translate(offsetX, offsetY, 0.f);

	// Compute the shadow space transform matrix by multiplying it with the view from light - used in the next step
	vsml->loadMatrix(VSMathLib::AUX0, shadowMatrix);
	vsml->multMatrix(VSMathLib::AUX0, vsml->get(VSMathLib::VIEW));

	// Load the projection matrix
	vsml->loadMatrix(VSMathLib::PROJECTION, shadowProjectionMatrix);

	glClear(GL_DEPTH_BUFFER_BIT);

	glUseProgram(shaderShadow.getProgramIndex());

	// Render offseted back faces
	glEnable(GL_POLYGON_OFFSET_FILL);
	glFrontFace(GL_CW);

	vsml->pushMatrix(VSMathLib::MODEL);
	// Move the torus so it casts nice shadow
	vsml->translate(0.f, 1.f, 0.f);

	// Draw the objects for shadow distances
	torusCaster.render();
	vsml->popMatrix(VSMathLib::MODEL);

	// ------------------------------------------------------------

	// Revert to prepare normal scene rendering
	glDisable(GL_POLYGON_OFFSET_FILL);
	glFrontFace(GL_CCW);

	// Go back to the window buffer
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	// Reset the viewport
	glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

	// Bind the generated shadow map
	glBindTexture(GL_TEXTURE_2D, shadowMap);

	// ------------------------------------------------------------

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera for the scene
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);

	vsml->translate(offsetX, offsetY, 0.f);

	// Load the projection matrix
	vsml->loadMatrix(VSMathLib::PROJECTION, cameraProjectionMatrix);

	// Transform light to camera space and send it to GLSL
	float res[4];
	vsml->multMatrixPoint(VSMathLib::VIEW, lightDir, res);
	vsml->normalize(res);
	shader.setBlockUniform("Lights", "l_dir", res);

	// Use shader for rendering with shadow
	glUseProgram(shader.getProgramIndex());

	shader.setUniform("time", time);

	vsml->pushMatrix(VSMathLib::MODEL);
	// Move the torus so it casts a nice shadow - same rotation as in shadow generation must be used!!!
	vsml->translate(0.f, 1.f, 0.f);

	torusCaster.render();
	vsml->popMatrix(VSMathLib::MODEL);

	// Move the sphere so that it gets the shadow
	vsml->translate(-5.f, -2.5f, -10.f);

	sphereReceiver.render();

	time += 1.f;

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;

		case 'm': glEnable(GL_MULTISAMPLE);
			break;

		case 'n': glDisable(GL_MULTISAMPLE);
			 break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for shadow usage
	shader.init();
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/useShadow.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/useShadow.frag");

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");

	shader.prepareProgram();
	printf("InfoLog for scene shader\n%s\n\n", shader.getAllInfoLogs().c_str());

	shader.setUniform("shadowUnit", 0); // Add sampler uniform

	// Shader for shadow generation
	shaderShadow.init();
	shaderShadow.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/genShadow.vert");
	shaderShadow.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/genShadow.frag");
	
	// Set semantics for the shader variables
	shaderShadow.setProgramOutput(0, "outputF");
	shaderShadow.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	shaderShadow.prepareProgram();
	printf("InfoLog for shadow shader\n%s\n\n", shaderShadow.getAllInfoLogs().c_str());

	return(shader.isProgramLinked() && shaderShadow.isProgramLinked());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	torusCaster.createTorus(0.5f, 1.0f, 50, 50);
	torusCaster.setMaterialBlockName("Materials");

	float tD[4] = {0.8f, 0.6f, 0.4f, 1.0f};
	torusCaster.setColor(VSResourceLib::DIFFUSE, tD);
	float tA[4] = {0.2f, 0.15f, 0.1f, 1.0f};
	torusCaster.setColor(VSResourceLib::AMBIENT, tA);

	sphereReceiver.createSphere(10.f, 50);
	sphereReceiver.setMaterialBlockName("Materials");

	float sD[4] = {0.8f, 0.9f, 0.8f, 1.0f};
	sphereReceiver.setColor(VSResourceLib::DIFFUSE, sD);
	float sA[4] = {0.2f, 0.2f, 0.2f, 1.0f};
	sphereReceiver.setColor(VSResourceLib::AMBIENT, sA);

	// ------------------------------------------------------------
	// Create textures for shadow map
	glGenTextures(1, &shadowMap);
	glBindTexture(GL_TEXTURE_2D, shadowMap);

	// Create a texture containing depths (floats).
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapSize, shadowMapSize, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);

	// Important: No interpolation.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// Set depth texture comparison function - lower value than texture means 1.0, higher means 0.0
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);

	// NOTICE: The texture will be bound during render
	glBindTexture(GL_TEXTURE_2D, 0);

	// Create framebuffers for shadow textures
	glGenFramebuffers(1, &shadowFramebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, shadowFramebuffer);

	// Do not use color buffers at all for this framebuffer
	glDrawBuffer(GL_NONE);
	glReadBuffer(GL_NONE);

	// Attach the depth texture.
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, shadowMap, 0);

	// Check correctness of shadow framebuffer
	if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
	{
		printf("GL_FRAMEBUFFER_COMPLETE error 0x%x", glCheckFramebufferStatus(GL_FRAMEBUFFER));

		return false;
	}

	// Set depth testing
	glClearDepth(1.0f);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// ------------------------------------------------------------

	// Some GL settings
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_MULTISAMPLE);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClearDepth(1.0f);

	// Needed when rendering the shadow map. This will avoid (some) artifacts.
	glPolygonOffset(1.0f, 0.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::VIEW, "m_view");
	vsml->setUniformName(VSMathLib::MODEL, "m_model");
	vsml->setUniformName(VSMathLib::NORMAL, "m_normal");
	vsml->setUniformName(VSMathLib::AUX0, "m_shadow"); // from world space to light's texture space
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Shadow Mapping");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}
